;; Generated package description from swiper.el  -*- no-byte-compile: t -*-
(define-package "swiper" "0.13.4.0.20230326.192004" "Isearch with an overview.  Oh, man!" '((emacs "24.5") (ivy "0.13.4")) :commit "87d5ecf12881a28377fd00fc6df67ad2ac06754d" :authors '(("Oleh Krehel" . "ohwoeowho@gmail.com")) :maintainer '("Oleh Krehel" . "ohwoeowho@gmail.com") :keywords '("matching") :url "https://github.com/abo-abo/swiper")
